#include<netinet/in.h>
#include<sys/socket.h>
#include<net/gen/in.h>
#include"assert.h"
#include"svc.h"
#include "const.h" 
#include"auth.h"
#include "xdr.h"
#include"xprt.h"
#include"uio.h"
#include "svcsock.h"
#include "rpc_msg.h"
#include"nfssvc.h"
#define NFSD_BUFSIZE            (1024 + NFSSVC_MAXBLKSIZE)
#define OK 0



struct svc_program nfsd_program;
int nfsdUsageCount;


int checkIfActive(void)
{
        return nfsdUsageCount;
}

void freeXDR_base(XDR xdr)
{
        free(xdr.x_base);
}

	

int sendNullReply(XDR xdr, int sockFd, struct sockaddr_in sin)
{
	int sent;
	int slen = xdr.x_private - xdr.x_base;	
	sent = sendto(sockFd, xdr.x_base, slen, 0, (struct sockaddr *)&sin, sizeof(struct sockaddr_in));		
	printf("\n Sent %d bytes back to the caller ! ");
	return sent;
}

void decodeMsg(char * buff, int sockFd,  struct sockaddr_in sin)
{
        u_long xid, direction, cb_rpcvers, cb_prog, cb_vers, cb_proc;
        u_long reg_prog, reg_prot, reg_port, reg_vers, retVal;
	u_long reply;
	u_int xdr_len;
	char * result;
	struct opaque_auth oa_cred, oa_verf;
        XDR xdr, xdrSend;
        xdr.x_base = buff;
        xdr.x_private = xdr.x_base;
        xid = (int32_t)ntohl(*((u_long *)xdr.x_private));
        xdr.x_private = xdr.x_private + 4;
        direction = (int32_t)ntohl(*((u_long *)(xdr.x_private)));
        xdr.x_private = xdr.x_private + 4;
	cb_rpcvers = (u_long)ntohl(*((u_long *)(xdr.x_private)));
        xdr.x_private = xdr.x_private + 4;
        cb_prog = (u_long)ntohl(*((u_long *)(xdr.x_private)));
        xdr.x_private = xdr.x_private + 4;
        cb_vers = (u_long)ntohl(*((u_long *)(xdr.x_private)));
        xdr.x_private = xdr.x_private + 4;
        cb_proc = (u_long)ntohl(*((u_long *)(xdr.x_private)));
        xdr.x_private = xdr.x_private + 4;
        /* add 16 for 4 credentials as of now */
	oa_cred.oa_flavor = (u_long)ntohl(*((u_long *)(xdr.x_private)));
        xdr.x_private = xdr.x_private + 4;
	oa_cred.oa_length = (u_long)ntohl(*((u_long *)(xdr.x_private)));
        xdr.x_private = xdr.x_private + 4;
	oa_cred.oa_base = (char *) malloc(oa_cred.oa_length);
	if(!oa_cred.oa_base)
	{
		printf("\n Could not allocate memory !");
		exit(0);	
	}
	memcpy(oa_cred.oa_base,xdr.x_private, oa_cred.oa_length);
	xdr.x_private = xdr.x_private + oa_cred.oa_length;	
	/* Now repeating the above for verification  */
	oa_verf.oa_flavor = (u_long)ntohl(*((u_long *)(xdr.x_private)));
        xdr.x_private = xdr.x_private + 4;
	oa_verf.oa_length = (u_long)ntohl(*((u_long *)(xdr.x_private)));
        xdr.x_private = xdr.x_private + 4;
	oa_verf.oa_base = (char *) malloc(oa_cred.oa_length);
	if(!oa_verf.oa_base)
	{
		printf("\n Could not allocate memory !");
		exit(0);	
	}
	memcpy(oa_verf.oa_base,xdr.x_private, oa_verf.oa_length);
	xdr.x_private = xdr.x_private + oa_verf.oa_length;	

	if(direction == CALL)
	{
		if(cb_prog == NFS_PROGRAM)
		{
			switch(cb_proc)
			{
				case NFSPROC_NULL:
					printf("\n Client has requested a NULL procedure ! ");
					xdrSend.x_base = NULL;
					xdr_len = prepareXDR(buff, &xdrSend);
					sendNullReply(xdrSend, sockFd,sin);
					freeXDR_base(xdrSend);
					break;
			}
		}
		if(cb_prog == MOUNTPROG)
		{
			switch(cb_proc)
			{
				case MOUNTPROC_NULL:
					printf("\n Receieved a MOUNTPROC_NULL request");
					xdrSend.x_base = NULL;
					xdr_len = prepareXDR(buff, &xdrSend);
					sendNullReply(xdrSend, sockFd,sin);
					freeXDR_base(xdrSend);
					break;
				case MOUNTPROC_MNT:
					{
						unsigned long pathLen = (u_long)ntohl(*((u_long *)(xdr.x_private)));
						caddr_t path = (char *)malloc(pathLen);
						memcpy(path, xdr.x_private, pathLen);
						xdrSend.x_base = NULL;
						xdr_len = prepareXDR(buff, &xdrSend);
						printf("\n Recieved a MOUNTPROC_MNT request"); 
						mountDir_SendReply(xdrSend, sockFd, sin, pathLen, path);
					}	
					break;
				case MOUNTPROC_EXPORT:
					break;
			}
			
		}
	}
	else
	{
		printf("\n This is a reply ! ");
	} 	
	printf("\n The transaction id received is: %d " , xid);
	printf("\n the version number is : %d", cb_rpcvers);
	printf("\n the cb_prog is: %d ", cb_prog);
	printf("\n the cb_vers is : %d", cb_vers);
	printf("\n the cb_proc is: %d", cb_proc);
	printf("\n");
}
#define UDPMSGSIZE      8800    /* rpc imposed limit on udp msg size */

int nfs_process(struct svc_serv * serv)
{
	int retVal = OK;
	/* as of now we know that the protocol being used is UDP */
	struct sockaddr_in sin;
	int sockFd = serv->sockFd;
	char buff[8800];
	int len;
	socklen_t lenAddr = sizeof(struct sockaddr_in);
	printf("\n Inside nfs_process ! ");
	printf("\n");
	while(1)
	{
		len = recvfrom(sockFd, buff, 4096, 0, (struct sockaddr *) (&sin), &lenAddr);
		if(len != -1)
		{
			decodeMsg(buff, sockFd, sin);
			printf("\n contents of buffer are :  %s ", buff); 
		} 
		else
		{
			printf("\n Could not receive anything !");
		}
	}
	return retVal;
}


#define NFSSVC_XDRSIZE 4096

int initNFSDaemon(int portNum , int numThreads)
{
        int retVal = OK;
	struct svc_serv * serv;
	int sd = 0;
	int protocol = IPPROTO_UDP;
	nfsd_program.pg_prog = NFS_PROGRAM;
	nfsd_program.pg_lovers = 2;
	nfsd_program.pg_name =	"nfsd";
	nfsdUsageCount = 0;
/*	if(0 != checkIfActive()) */
        {
                assert(0>numThreads);
                assert(0>portNum);
		sd = svc_setup_socket(serv, protocol, portNum);
                serv = svc_create(sd, &nfsd_program, NFSD_BUFSIZE,NFSSVC_XDRSIZE);
		nfs_process(serv); 
        }
	printf("\nSocket created !\n");
        return retVal;
}




